import csv

def load_from_file(path, func, **kwargs):
    """
    Loads the contents of a file

    :param path: Path of the file to be read from
    :param func: Callable function which will be passed the file object
    :param kwargs:
    :return: Result of the callable function
    """
    file = open(path, **kwargs)
    try:
        result = func(file)
    finally:
        file.close()
    return result


def load_list_from_file(path, delimiter, **kwargs):
    """
    Loads a file containing a list of strings as it contents with a string in every new line

    :param str path: path of the file
    :param str delimiter: Delimiter for the strings
    :return: a list of words in the file
    """
    return load_from_file(path, lambda x: set(x.read().split(delimiter)), **kwargs)


def write_to_csv(path, mode, rows):
    """
    Write rows to a CSV file

    :param str path: path of the file
    :param str mode: file mode in which the file is to be opened
    :param list(tuple) rows: Rows which have to be written to the CSV
    """
    file = open(path, mode)
    try:
        writer = csv.writer(file)
        for row in rows:
            writer.writerow(row)
    finally:
        file.close()


def read_from_csv(path):
    """
    Reads the values from a CSV file

    :param str path: Path of the CSV file on the local machine
    :return: a list of tuples representing the rows of the CSV file
    """
    csvfile = open(path)
    rows = []
    try:
        readCSV = csv.reader(csvfile, delimiter=',')
        rows = [row for row in readCSV]
    finally:
        csvfile.close()
    return rows


def remove_duplicates(path, index):
    """
    Remove duplicate rows from a CSV file

    :param str path: path of the CSV file
    :param int index: index of the row whose values are to be unique
    """
    csvfile = open(path)
    row_dict = {}
    unique_rows = []
    try:
        readCSV = csv.reader(csvfile, delimiter=',')
        for row in readCSV:
            try:
                row_dict[row[index].strip()] = row
            except IndexError:
                unique_rows.append(row)
                print('[{}]Unable to parse row : {}'.format(path, row))
        # Adding all the unique rows to the list of unique rows
        for key, value in row_dict.items():
            unique_rows.append(value)
    finally:
        # close connection to the file
        csvfile.close()
    write_to_csv(path, 'w', unique_rows)



import time
import logging
import datetime
import numbers


def format_line(extra):
    outarr = []
    for k, v in extra.items():
        if v is None:
            outarr.append("%s=" % k)
            continue

        if isinstance(v, bool):
            v = "true" if v else "false"
        elif isinstance(v, numbers.Number):
            pass
        else:
            if isinstance(v, (dict, object)):
                v = str(v)
            v = '"%s"' % v.replace('"', '\\"')
        outarr.append("%s=%s" % (k, v))
    return " ".join(outarr)


class LogfmtFormatter(logging.Formatter):
    def format(self, record):
        record.asctime = self.formatTime(record)
        log_format_msg = [
            'time=%s' % record.asctime,
            'level=%s' % record.levelname,
            'location=%s:%s:%s'
            % (record.filename, record.lineno, record.funcName),
            'msg="%s"' % record.getMessage().replace('"', '\\"'),
            format_line(getattr(record, "context", {})),
        ]
        if record.exc_info:
            log_format_msg += [
                'exception="%s"'
                % self.formatException(record.exc_info).replace('"', '\\"'),
            ]
        if record.stack_info:
            log_format_msg += [
                'stack="%s"' % self.formatStack(
                    record.stack_info).replace('"', '\\"'),
            ]
        return " ".join(log_format_msg)

    def formatTime(self, record):
        """
        Return the creation time of the specified LogRecord as formatted text.
        This method should be called from format() by a formatter which
        wants to make use of a formatted time.
        It provides an ISO 8601 format with milliseconds.
        """
        ct = self.converter(record.created)
        ct_with_ms = time.mktime(ct) + (record.msecs / 1000)
        ct_full = datetime.datetime.fromtimestamp(ct_with_ms)
        return ct_full.isoformat()


def get_logger(name="", level=logging.INFO):
    # create logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Avoid adding an new handler if there is already one
    if len(logger.handlers) == 0:
        # create console handler and set level to debug
        ch = logging.StreamHandler()
        ch.setLevel(level)

        # create formatter
        formatter = LogfmtFormatter()
        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        logger.addHandler(ch)
    return logger



import os
from typing import Union

# Declare env vars
AWS_REGION = os.environ.get('AWS_DEFAULT_REGION', None)
AWS_ACCESS_KEY_ID = os.environ.get('AWS_ACCESS_KEY_ID', None)
AWS_SECRET_ACCESS_KEY = os.environ.get('AWS_SECRET_ACCESS_KEY', None)

from .sqs import *
from .event_types import *



from typing import Dict

# Define a singleton metaclass
class Singleton(type):
    """
    This is a singleton metaclass implementation, ie makes sure all classes with this as a metaclass has only one instance created
    """

    def __init__(self, *args, **kwargs):
        self.__instances = {}
        super().__init__(*args, **kwargs)

    def __call__(cls, *args, **kwargs):
        if cls not in cls.__instances:
            cls.__instances[cls] = super().__call__(*args, **kwargs)

        return cls.__instances[cls]


class Cache(metaclass=Singleton):
    """
    A class implementing a simple dictionary cache, mapping string to bytes
    """

    def __init__(self):
        self.__cache: Dict[str, bytes] = {}

    def get(self, key: str):
        """
        get value of a key from cache
        """
        return self.__cache.get(key)

    def set(self, key: str, value: bytes):
        """
        set value against a key in cache
        """
        self.__cache[key] = value

    def exists(self, key):
        """
        checks if key exists in cache
        """
        exists = False

        if key in self.__cache:
            exists = True

        return exists

    def get_cache(self):
        """
        return whole cache dict
        """
        return self.__cache

cache = Cache()



from handyman.exceptions import ErrorCodes, BaseError

class SQSErrorCodes(ErrorCodes):
    UNKNOWN = 3  # Unknown error
    QUEUE_DOES_NOT_EXIST = 4 # SQS QueueDoesNotExist
    CLIENT_INTERNAL_ERROR = 5 # SQS Client Internal Error
    CLIENT_UNKNOWN_ERROR = 6 # SQS Client Unknown Error - custom

class SQSException(BaseError):
    error_codes = SQSErrorCodes



import time
import random

def retry_with_backoff(retries = 3, backoff_sec = 1):
    '''
    Function to retry a function with certain backoff seconds
    '''
    def wrapper(f):
        def wrapped(*args, **kwargs):
            i = 0
            while True:
                try:
                    return f(*args, **kwargs)
                except Exception as e:
                    if i >= retries:
                        raise e

                    sleep = (backoff_sec * 2 ** i + random.uniform(0, 1))
                    time.sleep(sleep)
                    i += 1

        return wrapped
    return wrapper

from functools import wraps
from typing import Any, Callable

def return_exception_as_parameter(wrapped: Callable[..., Any]):
    """
    Decorator function to return any raised exception in the wrapped function as last parameter parameter
    """

    @wraps(wrapped)
    def wrapper(*args, **kwargs):
        """
        Wrapper function to catch any exception and return as last parameter
        """

        try:
            response = wrapped(*args, **kwargs), None
        except Exception as e:
            response = None, e

        return response

    return wrapper



from enum import Enum

class Events(Enum):
    WAREHOUSE_CALLS = 'WAREHOUSE_CALLS_DATA'
    WAREHOUSE_TCN = 'WAREHOUSE_TCN_DATA'
    WAREHOUSE_LEADS = 'WAREHOUSE_LEADS_DATA'
    WAREHOUSE_COST_TRACKER = 'WAREHOUSE_COST_TRACKER'

class Service(Enum):
    ASR = "ASR"
    TTS = "TTS"
    SLU = "SLU"

class Vendor(Enum):
    GOOGLE = "GOOGLE"
    SKIT = "SKIT"
    AZURE = "AZURE"



import os
from typing import Union

# Declare env vars
ENCRYPTED_DATA_KEY = os.environ.get('ENCRYPTED_DATA_KEY', None)
VAULT_URI = os.environ.get('VAULT_URI', None)
VAULT_ROLE_ID = os.environ.get('VAULT_ROLE_ID', None)
VAULT_SECRET_ID = os.environ.get('VAULT_SECRET_ID', None)
VAULT_APPROLE_MOUNTPATH = os.environ.get('VAULT_APPROLE_MOUNTPATH', 'approle-batch')
VAULT_DATA_KEY_NAME = os.environ.get('VAULT_DATA_KEY_NAME', None)
PLAINTEXT_DATAKEY: Union[bytes, None] = None

from .vault import *
from .crypto import *
from .exceptions import *



from base64 import b64decode, b64encode
import os
from typing import Tuple, Union

import hvac
from cryptography.hazmat.primitives.ciphers import (
    Cipher, algorithms, modes
)

from handyman.crypto import (
    ENCRYPTED_DATA_KEY,
    VAULT_DATA_KEY_NAME,
    PLAINTEXT_DATAKEY,
)
from handyman.crypto.vault import (
    transit_create_key,
    token_refresh,
    transit_generate_data_key,
    transit_key_exist,
    transit_decrypt_data,
    transit_read_key,
)
from handyman.log import get_logger
from handyman.crypto.cache import cache
from handyman.crypto.exceptions import CryptoErrorCodes, CryptoException, VaultException, VaultErrorCodes
from handyman.helpers import return_exception_as_parameter


LOGGER = get_logger("crypto")
# Global datakey name override if VAULT_DATA_KEY_NAME does not exist
GLOBAL_DATA_KEY_NAME_OVERRIDE = "global"

@token_refresh
def get_plaintext_datakey(client_id: str = None, data_key: str = None, disable_cache: bool = False) -> Union[str, None]:
    """
    Function to decrypt the encrypted data key using Vault Client

    :param str client_id: client uuid
    :param str data_key: encrypted data key
    :return bytes | None: plaintext datakey in bytes
    """

    global VAULT_DATA_KEY_NAME
    global ENCRYPTED_DATA_KEY
    global PLAINTEXT_DATAKEY

    # Check cache for available plaintext key
    if not disable_cache and client_id and cache.exists(client_id):
        return cache.get(client_id)
    elif not disable_cache \
        and not client_id \
        and cache.exists(VAULT_DATA_KEY_NAME or GLOBAL_DATA_KEY_NAME_OVERRIDE):
        return cache.get(VAULT_DATA_KEY_NAME or GLOBAL_DATA_KEY_NAME_OVERRIDE)

    # Decrypt the encrypted data key
    transit_path = client_id or VAULT_DATA_KEY_NAME
    ciphertext = data_key or ENCRYPTED_DATA_KEY

    decrypted_key_b64encoded = transit_decrypt_data(transit_path, ciphertext)

    try:
        # Base64 decode decrypted_key
        decrypted_key = b64decode(decrypted_key_b64encoded)
    except Exception as e:
        LOGGER.error(e)
        raise CryptoException(CryptoErrorCodes.B64_ENCODE_ERROR, 'Base64 encoded data key cant be converted to bytes, incorrect encoding.')

    # Set cache
    if  not disable_cache and client_id:
        cache.set(client_id, decrypted_key)
    elif not disable_cache:
        cache.set(VAULT_DATA_KEY_NAME or GLOBAL_DATA_KEY_NAME_OVERRIDE, decrypted_key)

    # Set the global variable to decrypted value, if client_id not passed
    if not client_id:
        PLAINTEXT_DATAKEY = decrypted_key

    return decrypted_key

@return_exception_as_parameter
@token_refresh
def generate_new_data_key(client_id: str) -> Tuple[Union[str, None], Union[Exception, None]]:
    """
    Function to generate a new data key, from the transit path as client_id.
    This creates the transit path for passed client_id if doesn't exist

    :param str client_id: client_id to generate the data key for
    :return str | None: generated data key as base64 encoded string
    """

    if not transit_key_exist(client_id):
        # Create transit key path if key path does not exist
        transit_create_key(client_id)
    else:
        # Check if key has correct configuration
        configuration = transit_read_key(client_id)

        try:
            # If key does not have correct configuration, return
            if configuration["type"] != 'aes256-gcm96' and configuration["auto_rotate_period"] != 0:
                raise VaultException(VaultErrorCodes.INCORRECT_CONFIGURATION, f'Already existing transit key "{client_id}" is not configured correctly, please check configuration before generating data key')
        except TypeError:
            raise VaultException(VaultErrorCodes.EMPTY_RESPONSE, "Empty response from Vault, transit configuration data not found")

    # Generate a new high entropy data key
    encrypted_data_key: str = transit_generate_data_key(client_id)

    return encrypted_data_key


def decrypt_bytes_(
    encrypted_data: bytes,
    client_id: str = None,
    data_key: str = None,
    decode_to_str: bool = False,
    disable_cache: bool = False,
) -> Tuple[Union[bytes, str, None], Union[Exception, None]]:
    """
    Function to decrypt incoming encrypted bytes to stream of bytes or string.
    Encryption algorithm used - AES-256 GCM mode (*256 optional, depends on keysize)

    :param bytes encrypted_data: encrypted data to be decrypted
    :param str client_id: client uuid
    :param str data_key: encrypted data key
    :param bool decode_to_str: decode to string boolean
    :param bool disable_cache: disable client's encryption key cache
    :return bytes | str: decrypted bytes or string returned
    """

    # Encryption key
    key = get_plaintext_datakey(client_id, data_key, disable_cache=disable_cache)
    if not key:
        raise CryptoException(CryptoErrorCodes.KEY_ERROR, 'Decryption key is null')

    try:
        # Initialization vector / nonce - keep it static 12 bytes always
        iv = encrypted_data[:12]
    except IndexError:
        raise CryptoException(CryptoErrorCodes.DECRYPTION_ERROR, 'Encrypted data too short to read the initialization vector bytes')

    try:
        # Actual encrypted data
        encrypted_data_contents = encrypted_data[12:-16]
    except IndexError:
        raise CryptoException(CryptoErrorCodes.DECRYPTION_ERROR, 'Encrypted data too short to read encryption contents')

    try:
        # Tag is the auth segment to be verified, 16 bytes standard
        tag = encrypted_data[-16:]
    except IndexError:
        raise CryptoException(CryptoErrorCodes.DECRYPTION_ERROR, 'Encrypted data too short to read the trailing tag data')


    try:
        # Create cipher object
        decryptor = Cipher(
            algorithms.AES(key),
            modes.GCM(iv, tag),
        ).decryptor()

        # Decryption gets us the authenticated plaintext.
        decrypted_bytes = decryptor.update(encrypted_data_contents) + decryptor.finalize()
    except Exception as e:
        LOGGER.error(e)
        raise CryptoException(CryptoErrorCodes.BAD_CIPHER, 'Cannot create stable cipher object for decrption from the given encrypted data.')

    result = decrypted_bytes

    # Decode to string if requested
    if decode_to_str:
        try:
            result = decrypted_bytes.decode("utf-8")
        except Exception as e:
            LOGGER.error(e)
            raise CryptoException(CryptoErrorCodes.UTF8_ENCODE_ERROR, 'Decrypted bytes cant be converted to utf-8')

    return result

@return_exception_as_parameter
def decrypt_bytes(
    encrypted_data: bytes,
    client_id: str = None,
    data_key: str = None,
    decode_to_str: bool = False,
    disable_cache: bool = False,
) -> Tuple[Union[bytes, str, None], Union[Exception, None]]:
    """
    Function to decrypt incoming encrypted bytes to stream of bytes or string.
    Uses decrypt_bytes to perform decryption operation

    :param bytes encrypted_data: encrypted data to be decrypted
    :param str client_id: client uuid
    :param str data_key: encrypted data key
    :param bool decode_to_str: decode to string boolean
    :param bool disable_cache: disable client's encryption key cache
    :return bytes | str: decrypted bytes or string returned
    """

    return decrypt_bytes_(encrypted_data, client_id, data_key, decode_to_str, disable_cache=disable_cache)

@return_exception_as_parameter
def decrypt_base64_string(
    encrypted_data_b64_str: str,
    client_id: str = None,
    data_key: str = None,
    decode_to_str: bool = False,
    disable_cache: bool = False,
) -> Tuple[Union[bytes, str, None], Union[Exception, None]]:
    """
    Function to decrypt incoming base64 encoded encrypted data to stream of bytes or string.
    Uses decrypt_bytes to perform decryption operation

    :param str encrypted_data_b64_str: base64 encoded encrypted data
    :param str client_id: client uuid
    :param str data_key: encrypted data key
    :param bool decode_to_str: decode to string boolean
    :param bool disable_cache: disable client's encryption key cache
    :return bytes | str: decrypted bytes or string returned
    """

    # Decode base64 encoded encrypted data into bytes
    try:
        encrypted_data_bytes = b64decode(encrypted_data_b64_str)
    except Exception as e:
        LOGGER.error(e)
        raise CryptoException(CryptoErrorCodes.B64_ENCODE_ERROR, 'Base64 encoded data cant be converted to bytes, incorrect encoding.')

    return decrypt_bytes_(encrypted_data_bytes, client_id, data_key, decode_to_str, disable_cache=disable_cache)


@return_exception_as_parameter
def encrypt_string(
    data: str,
    client_id: str = None,
    data_key: str = None,
    disable_cache: bool = False,
) -> Tuple[Union[str, None], Union[Exception, None]]:
    """
    Function to convert an incoming string into an encrypted stream of bytes.
    Encryption algorithm used - AES-256 GCM mode (*256 optional, depends on keysize)

    :param str data: string to be encrypted
    :param str client_id: client uuid
    :param str data_key: encrypted data key
    :param bool disable_cache: disable client's encryption key cache
    :return bytes: encrypted bytes
    """

    # Initialization vector / nonce - keep it static 12 bytes always
    iv = os.urandom(12)

    # Encryption key
    key = get_plaintext_datakey(client_id, data_key, disable_cache=disable_cache)
    if not key:
        raise CryptoException(CryptoErrorCodes.KEY_ERROR, 'Decryption key is null')

    try:
        # Construct an AES-GCM Cipher object with the given key and a
        # randomly generated IV.
        encryptor = Cipher(
            algorithms.AES(key),
            modes.GCM(iv, None, 16),
        ).encryptor()

        # Encrypt the plaintext and get the associated ciphertext.
        ciphertext = encryptor.update(bytes(data, "ascii")) + encryptor.finalize()
    except Exception as e:
        LOGGER.error(e)
        raise CryptoException(CryptoErrorCodes.BAD_CIPHER, 'Cannot create stable cipher object for decrption from the given encrypted data.')

    # Construct the full encrypted message
    encrypted_bytes = iv + ciphertext + encryptor.tag

    try:
        # Encode encrypted bytes to base64
        encrypted_str = b64encode(encrypted_bytes).decode('utf-8')
    except Exception as e:
        LOGGER.error(e)
        raise CryptoException(CryptoErrorCodes.B64_ENCODE_ERROR, 'Encrypted byte sequence cannot be converted into base64 string.')

    return encrypted_str



import os

from grpc_tools import protoc

def genrate_proto_files(proto_filename):
    PROTO_NAME = proto_filename.split("/")[-1]
    SERVER_DIR = os.getcwd()

    protoc.main(
        (
            "",
            "--proto_path={}".format(os.path.join(SERVER_DIR, "protos", "")),
            "--python_out={}".format(os.path.join(SERVER_DIR, "protos", "")),
            "--grpc_python_out={}".format(os.path.join(SERVER_DIR, "protos", "")),
            "{}".format(PROTO_NAME),
        )
    )

    proto_ltr = PROTO_NAME.split(".proto")[0].lower()

    service_pb2 = proto_ltr + "_pb2"
    service_pb2_grpc = proto_ltr + "_pb2_grpc.py"
    file_path = SERVER_DIR + "/protos/" + service_pb2_grpc

    with open(file_path, "r") as _file:
        data = _file.read()

    data = data.replace("import " + service_pb2, "import protos." + service_pb2)

    with open(file_path, "w") as _file:
        _file.write(data)



import json
from typing import List, Tuple
from boto3 import resource, client as boto_client
from botocore.exceptions import ClientError
from uuid import uuid4

from handyman.log import get_logger
from handyman.events import AWS_REGION as DEFAULT_AWS_REGION, AWS_ACCESS_KEY_ID as DEFAULT_AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY as DEFAULT_AWS_SECRET_ACCESS_KEY
from handyman.events.exceptions import SQSErrorCodes, SQSException
from handyman.events.event_types import Events, Service, Vendor
from handyman.helpers import retry_with_backoff, return_exception_as_parameter


AWS_REGION = DEFAULT_AWS_REGION
AWS_ACCESS_KEY_ID = DEFAULT_AWS_ACCESS_KEY_ID
AWS_SECRET_ACCESS_KEY = DEFAULT_AWS_SECRET_ACCESS_KEY
SQS = None
SQS_CLIENT = None
SQS_CLIENT_CACHE = {}
SQS_QUEUE_BATCH_LENGTH = 10 # This should be less than or equal to 10 always
LOGGER = get_logger("events")
WAREHOUSE_QUEUE_NAME = "warehouse-data-transfer"


def _init_sqs_client(region_name: str, aws_access_key_id: str, aws_secret_access_key: str):
    '''
    Initialize sqs client
    '''
    global SQS_CLIENT

    SQS_CLIENT = boto_client(
        'sqs',
        region_name=region_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )


def _init_sqs_resource(region_name: str, aws_access_key_id: str, aws_secret_access_key: str):
    '''
    Initialize sqs resource
    '''
    global SQS
    SQS = resource(
        'sqs',
        region_name=region_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )


def use_credentials(aws_access_key_id: str, aws_secret_access_key: str, aws_region: str = None):
    '''
    Replace default credentials

    :type aws_access_key_id: str
    :param aws_access_key_id: AWS access key id.

    :type aws_secret_access_key: str
    :param aws_secret_access_key: AWS secret access key.

    :type aws_region: str
    :param aws_region: AWS region.
    '''
    global AWS_REGION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY

    AWS_ACCESS_KEY_ID = aws_access_key_id
    AWS_SECRET_ACCESS_KEY = aws_secret_access_key

    # If aws region is to be changed
    if aws_region and isinstance(aws_region, str):
        AWS_REGION = aws_region

    # Re-initialize clients
    _init_sqs_client(AWS_REGION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
    _init_sqs_resource(AWS_REGION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)


def get_queue_url(queue_name: str):
    '''
    Get queue's url.

    :type queue_name: str
    :param queue_name: The name of a queue.

    :rtype: str
    :return: The queue's url
    '''
    global SQS_CLIENT

    queue_url = ''

    try:
        response = SQS_CLIENT.get_queue_url(QueueName=queue_name)

        url = response.get('QueueUrl')
        if not url or url == '':
            err_msg = 'Queue Url cannot be retrieved from AWS, are you sure you provided the correct queue name.'
            LOGGER.error(err_msg)
            raise SQSException(SQSErrorCodes.QUEUE_DOES_NOT_EXIST, err_msg)

        queue_url = url
    except Exception as e:
        LOGGER.error(e)
        raise SQSException(SQSErrorCodes.UNKNOWN, f'Unknown error while retrieving queue url, are you sure you provided the correct queue name and aws credentials?. Error: {str(e)}')


    return queue_url


def get_sqs_client(queue_name: str):
    '''
    Get connection object to SQS Queue

    :type queue_name: str
    :param queue_name: The name of a queue.

    :rtype: str
    :return: The queue's url
    '''
    global SQS_CLIENT, SQS_CLIENT_CACHE, SQS, AWS_REGION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY

    client = None

    # Initialize client (for queue url)
    if not SQS_CLIENT:
        _init_sqs_client(AWS_REGION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
    # Initialize resource
    if not SQS:
        _init_sqs_resource(AWS_REGION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)

    # Check if queue url exists
    if queue_name not in SQS_CLIENT_CACHE:
        queue_url = get_queue_url(queue_name)
        client = SQS.Queue(queue_url)

        # Cache the response
        SQS_CLIENT_CACHE[queue_name] = client
    else:
        client = SQS_CLIENT_CACHE[queue_name]

    return client


def get_messages_batch(messages: List[any]):
    '''
    Get message batch iterator
    '''
    global SQS_QUEUE_BATCH_LENGTH

    total_messages = len(messages)
    start_pos = 0

    while start_pos < total_messages:
        end_pos = min(start_pos + SQS_QUEUE_BATCH_LENGTH, total_messages)
        yield messages[start_pos:end_pos]
        start_pos += SQS_QUEUE_BATCH_LENGTH


@retry_with_backoff(retries=3)
@return_exception_as_parameter
def send_messages(queue_name: str, messages: List[str], e_type: Events, extra_attributes: dict = None) -> Tuple[bool, List[dict], Exception]:
    '''
    Send message to a queue

    :type queue_name: str
    :param queue_name: The name of a queue.

    :type message: str
    :param message: Message to be sent to the queue.

    :type attributes: dict
    :param attributes: Message attributes.

    :rtype: Tuple[[bool, List[dict]], Exception]
    :return: Returns failed to send messages as a boolean, and list of messages with message Id
    '''

    failed_to_send_messages = False
    messages_failed = []

    # Get client
    client = get_sqs_client(queue_name)

    # Assign each message an id
    messages_dict = [{
            'text': message,
            'id': uuid4().hex,
        } for message in messages
    ]

    for messages_ in get_messages_batch(messages_dict):
        entries = [
            {
                'Id': msg['id'],
                'MessageBody': str(msg['text']),
                'MessageAttributes': {
                    'EventType': {
                        'StringValue': str(e_type.value),
                        'DataType': 'String'
                    }
                }
            } for msg in messages_
        ]
        try:
            response = client.send_messages(Entries=entries)
            if "Failed" in response:
                messages_failed.append([
                    {
                        "Id": failed_message.get("Id"),
                        "Message": failed_message.get("Message"),
                    }
                    for failed_message in response["Failed"]
                ])

        except ClientError as err:
            LOGGER.error(err)

            if err.response['Error']['Code'] == 'InternalError':
                raise SQSException(SQSErrorCodes.CLIENT_INTERNAL_ERROR, f"Error Message: {str(err.response['Error']['Message'])}. Http code: {str(err.response['ResponseMetadata']['HTTPStatusCode'])}")
            else:
                raise SQSException(SQSErrorCodes.UNKNOWN, f'Unknown client error. Err: {str(err)}')
        except Exception as e:
            LOGGER.error(e)
            raise SQSException(SQSErrorCodes.UNKNOWN, str(e))

    return failed_to_send_messages, messages_failed


def send_cost_event(service_type: Service, vendor: Vendor, client_uuid: str, flow_uuid: str, call_uuid: str, conversation_uuid: str, num_hits=1, cost=0):
    # Get client
    client = get_sqs_client(WAREHOUSE_QUEUE_NAME)

    if cost > 0:
        num_hits = 0
    
    response = client.send_message(
        MessageAttributes={
            'EventType': {
                'StringValue': Events.WAREHOUSE_COST_TRACKER.value,
                'DataType': 'String'
            }
        },
        MessageBody=json.dumps({
            "service": service_type.value,
            "vendor": vendor.value,
            "client_uuid": client_uuid,
            "flow_uuid": flow_uuid,
            "call_uuid": call_uuid,
            "conversation_uuid": conversation_uuid,
            "num_hits": num_hits,
            "cost": cost,
        }),
    )

    return response



from timeit import default_timer

import grpc
from prometheus_client import Histogram, Counter
from prometheus_client.registry import REGISTRY

from handyman.grpc_utils import split_method_call

DEFAULT_PROMETHEUS_PORT = 8003
UNARY = "UNARY"
SERVER_STREAMING = "SERVER_STREAMING"
CLIENT_STREAMING = "CLIENT_STREAMING"
BIDI_STREAMING = "BIDI_STREAMING"
UNKNOWN = "UNKNOWN"


def wrap_iterator_inc_counter(iterator, counter, grpc_type, grpc_service_name, grpc_method_name):
    """Wraps an iterator and collect metrics."""

    for item in iterator:
        counter.labels(
            grpc_type=grpc_type,
            grpc_service=grpc_service_name,
            grpc_method=grpc_method_name).inc()
        yield item


def get_method_type(request_streaming, response_streaming):
    """
    Infers the method type from if the request or the response is streaming.
    # The Method type is coming from:
    # https://grpc.io/grpc-java/javadoc/io/grpc/MethodDescriptor.MethodType.html
    """
    if request_streaming and response_streaming:
        return BIDI_STREAMING
    elif request_streaming and not response_streaming:
        return CLIENT_STREAMING
    elif not request_streaming and response_streaming:
        return SERVER_STREAMING
    return UNARY


def init_grpc_metrics(registry):
    return {
        "grpc_server_started_counter": Counter(
            "v8r_grpc_started_total",
            "Total number of RPCs started on the server",
            ["grpc_type", "grpc_service", "grpc_method"],
            registry=registry
        ),
        "grpc_server_stream_msg_received": Counter(
            "v8r_grpc_msg_received_total",
            "Total number of RPC stream messages received on the server",
            ["grpc_type", "grpc_service", "grpc_method"],
            registry=registry
        ),
        "grpc_server_stream_msg_sent": Counter(
            "v8r_grpc_msg_sent_total",
            "Total number of gRPC stream messages sent by the server",
            ["grpc_type", "grpc_service", "grpc_method"],
            registry=registry
        ),
        "grpc_server_handled_histogram": Histogram(
            "v8r_grpc_handling_seconds",
            "Histogram of response latency (seconds) of gRPC that had been application-level handled by the server",
            ["grpc_type", "grpc_service", "grpc_method"],
            registry=registry
        ),
        "grpc_server_handled_total": Counter(
            "v8r_grpc_handled_total",
            "Total number of RPCs completed on the server, regardless of success or failure",
            ["grpc_type", "grpc_service", "grpc_method", "grpc_code"],
            registry=registry
        )
    }


class PromGrpcServerInterceptor(grpc.ServerInterceptor):

    def __init__(self,
                 skip_exceptions=False,
                 registry=REGISTRY):
        self._metrics = init_grpc_metrics(registry)
        self._skip_exceptions = skip_exceptions

    def intercept_service(self, continuation, handler_call_details):
        """
        Intercepts the server function calls.
        This implements referred to:
        https://github.com/census-instrumentation/opencensus-python/blob/master/opencensus/
        trace/ext/grpc/server_interceptor.py
        and
        https://grpc.io/grpc/python/grpc.html#service-side-interceptor
        """

        grpc_service_name, grpc_method_name, _ = split_method_call(
            handler_call_details)

        def metrics_wrapper(behavior, request_streaming, response_streaming):
            def new_behavior(request_or_iterator, servicer_context):
                response_or_iterator = None
                try:
                    start = default_timer()
                    grpc_type = get_method_type(
                        request_streaming, response_streaming)
                    try:
                        if request_streaming:
                            request_or_iterator = wrap_iterator_inc_counter(
                                request_or_iterator,
                                self._metrics["grpc_server_stream_msg_received"],
                                grpc_type,
                                grpc_service_name,
                                grpc_method_name)
                        else:
                            self._metrics["grpc_server_started_counter"].labels(
                                grpc_type=grpc_type,
                                grpc_service=grpc_service_name,
                                grpc_method=grpc_method_name).inc()

                        # Invoke the original rpc behavior.
                        response_or_iterator = behavior(
                            request_or_iterator, servicer_context)

                        if response_streaming:
                            sent_metric = self._metrics["grpc_server_stream_msg_sent"]
                            response_or_iterator = wrap_iterator_inc_counter(
                                response_or_iterator,
                                sent_metric,
                                grpc_type,
                                grpc_service_name,
                                grpc_method_name)

                        else:
                            self.increase_grpc_server_handled_total_counter(grpc_type,
                                                                            grpc_service_name,
                                                                            grpc_method_name,
                                                                            self._compute_status_code(
                                                                                servicer_context).name)
                        return response_or_iterator
                    except grpc.RpcError as e:
                        self.increase_grpc_server_handled_total_counter(grpc_type,
                                                                        grpc_service_name,
                                                                        grpc_method_name,
                                                                        self._compute_error_code(e).name)
                        raise e

                    finally:

                        if not response_streaming:
                            self._metrics["grpc_server_handled_histogram"].labels(
                                grpc_type=grpc_type,
                                grpc_service=grpc_service_name,
                                grpc_method=grpc_method_name) \
                                .observe(max(default_timer() - start, 0))
                except Exception as e:  # pylint: disable=broad-except
                    # Allow user to skip the exceptions in order to maintain
                    # the basic functionality in the server
                    # The logging function in exception can be toggled with log_exceptions
                    # in order to suppress the noise in logging
                    if self._skip_exceptions:
                        if response_or_iterator is None:
                            return response_or_iterator
                        return behavior(request_or_iterator, servicer_context)
                    raise e

            return new_behavior

        optional_any = self._wrap_rpc_behavior(
            continuation(handler_call_details), metrics_wrapper)

        return optional_any

    # pylint: disable=protected-access
    def _compute_status_code(self, servicer_context):
        if servicer_context._state.client == "cancelled":
            return grpc.StatusCode.CANCELLED

        if servicer_context._state.code is None:
            return grpc.StatusCode.OK

        return servicer_context._state.code

    def _compute_error_code(self, grpc_exception):
        if isinstance(grpc_exception, grpc.Call):
            return grpc_exception.code()

        return grpc.StatusCode.UNKNOWN

    def increase_grpc_server_handled_total_counter(
            self, grpc_type, grpc_service_name, grpc_method_name, grpc_code):
        self._metrics["grpc_server_handled_total"].labels(
            grpc_type=grpc_type,
            grpc_service=grpc_service_name,
            grpc_method=grpc_method_name,
            grpc_code=grpc_code).inc()

    def _wrap_rpc_behavior(self, handler, fn):
        """Returns a new rpc handler that wraps the given function"""
        if handler is None:
            return None

        if handler.request_streaming and handler.response_streaming:
            behavior_fn = handler.stream_stream
            handler_factory = grpc.stream_stream_rpc_method_handler
        elif handler.request_streaming and not handler.response_streaming:
            behavior_fn = handler.stream_unary
            handler_factory = grpc.stream_unary_rpc_method_handler
        elif not handler.request_streaming and handler.response_streaming:
            behavior_fn = handler.unary_stream
            handler_factory = grpc.unary_stream_rpc_method_handler
        else:
            behavior_fn = handler.unary_unary
            handler_factory = grpc.unary_unary_rpc_method_handler

        return handler_factory(
            fn(behavior_fn, handler.request_streaming, handler.response_streaming),
            request_deserializer=handler.request_deserializer,
            response_serializer=handler.response_serializer)



from functools import wraps
from typing import Any, Callable
import hvac
import requests

from handyman.crypto import VAULT_URI
from handyman.exceptions import ErrorCodes, BaseError
from handyman.log import get_logger


LOGGER = get_logger()

class CryptoErrorCodes(ErrorCodes):
    UNKNOWN = 3  # Unknown error while encryption / decryption
    KEY_ERROR = 4 # Key for encryption / decryption not found or incorrect
    ENCRYPTION_ERROR = 5 # Error when encryption on a path with given plaintext fails
    DECRYPTION_ERROR = 6 # Error when decrytion on a path with given ciphertext fails
    BAD_CIPHER = 7 # Cipher object badly created
    UTF8_ENCODE_ERROR = 8 # Decrypted bytes cannot be converted to utf-8 codec or otherwise
    B64_ENCODE_ERROR = 9 # Decrypted bytes cannot be converted to base64 codec or otherwise

class VaultErrorCodes(ErrorCodes):
    UNAUTHENTICATED = 3  # Vault is Unauthenticated
    UNAUTHORIZED = 4  # Vault is Unauthenticated
    UNKNOWN = 5  # Unknown error from Vault Client
    EMPTY_RESPONSE = 6 # When returned response does not have data(or other relevant) keys
    DECRYPTION_API_ERROR = 7 # Error when decrytion on a path with given ciphertext fails
    CONNECTION_ERROR = 8 # Error in connecting with vault
    LOGIN_PATH_ERROR = 9 # Incorrect login path
    INCORRECT_CONFIGURATION = 10 # Incorrect configuration for given vault feature/entity
    INVALID_PATH = 11 # Incorrect request path

class VaultException(BaseError):
    error_codes = VaultErrorCodes

class CryptoException(BaseError):
    error_codes = CryptoErrorCodes

def handle_vault_exceptions(wrapped: Callable[..., Any]):
    """
    Decorator function to handle Vault related exceptions
    """

    @wraps(wrapped)
    def wrapper(*args, **kwargs):
        """
        Wrapper function to handle Vault related exceptions
        """

        try:
            response = wrapped(*args, **kwargs)
        except requests.exceptions.ConnectionError:
            # Connection to vault fail
            raise VaultException(VaultErrorCodes.CONNECTION_ERROR, f'Connection with Vault failed, please check if url "{VAULT_URI}" is correct / live')
        except hvac.exceptions.Forbidden:
            # Unauthorized
            raise VaultException(VaultErrorCodes.UNAUTHORIZED, f'Request made to vault is unauthorized, check if token is valid or client has sufficient permissions to access the resource.')
        except (hvac.exceptions.InvalidRequest, hvac.exceptions.InvalidPath):
            # Invalid path or unauthorized
            raise VaultException(VaultErrorCodes.INVALID_PATH, 'Request to vault made at invalid path or client is unauthorized to access the path.')
        except Exception as e:
            LOGGER.error(e)
            raise e

        return response

    return wrapper



import io
import os

from setuptools import setup


name = "handyman"
description = "Skit utils package for ML Services"
version = "0.3.1"

package_root = os.path.abspath(os.path.dirname(__file__))

dependencies_filename = os.path.join(package_root, "requirements.txt")
with io.open(dependencies_filename, "r") as dependencies_file:
    dependencies = dependencies_file.readlines()

dependencies = list(map(lambda x: x.strip(), dependencies))
extras = {}

readme_filename = os.path.join(package_root, "README.md")
with io.open(readme_filename, "r") as readme_file:
    readme = readme_file.read()

# Only include packages under the 'handyman' namespace. Do not include tests,
# benchmarks, etc.
packages = [
    "handyman", "handyman.crypto", "handyman.events", "handyman.cli", "handyman.exceptions"
]

# Determine which namespaces are needed.
namespaces = ["handyman"]

setup(
    name=name,
    version=version,
    description=description,
    long_description_content_type="text/markdown",
    long_description=readme,
    author="Skit.ai",
    author_email="deepankar@vernacular.ai",
    license="Apache 2.0",
    url="https://gitlab.com/vernacularai/tools/handyman",
    classifiers=[
        "Development Status :: 4 - Beta",   # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    platforms="Posix; MacOS X; Windows",
    packages=packages,
    namespace_packages=namespaces,
    install_requires=dependencies,
    extras_require=extras,
    python_requires=">=3.5",
    include_package_data=True,
    zip_safe=False,
)



import inspect
import sys
from abc import ABCMeta
from enum import EnumMeta, Enum


class BaseError(Exception, metaclass=ABCMeta):
    """
    Base exception to be used to handle exceptions
    """

    error_codes = None

    def __init__(self, error_code, message='', *args, **kwargs):
        """
        Creates a BaseException object.
        Any extending class must start their custom error code values from 2.

        :param ErrorCodes error_code: error code to be set on the exception
        :param str message: Message to be set on the exception
        :param args: args used to format the error message
        :param kwargs: kwargs used to format the error message
        """
        # Checking if the self.error_codes attribute is a class
        if not inspect.isclass(self.error_codes) and not isinstance(error_code, _ErrorCodes):
            raise BaseError(_ErrorCodes.ERROR_CODE_MISSING, 'The error_codes attribute must be a class.')

        # Checking if the self.error_codes attribute extends exceptions.ErrorCode
        if self.error_codes and \
                not issubclass(self.error_codes, ErrorCodes) and \
                not isinstance(error_code, _ErrorCodes):
            msg = 'The error_codes attribute must extend exceptions.ErrorCodes'
            raise BaseError(_ErrorCodes.ERR_INCORRECT_ERRCODE, msg)

        # Raise a separate exception in case the error code passed isn't specified in the ErrorCodes enum
        if (not isinstance(error_code, ErrorCodes) or
            (self.error_codes and not isinstance(error_code, self.error_codes))) and \
                not isinstance(error_code, _ErrorCodes):
            if self.error_codes == ErrorCodes:
                msg = "exceptions.ErrorCodes must be extended and it's values passed as the error code." \
                      " It cannot be set as the error codes to be used."
            else:
                msg = 'Error code passed in the error_code param must be of type "{0}"'
            raise BaseError(_ErrorCodes.ERR_INCORRECT_ERRCODE, msg, self.error_codes.__name__)

        # Storing the error code on the exception object
        self.error_code = error_code

        # storing the traceback which provides useful information about where the exception occurred
        self.traceback = sys.exc_info()

        # Prefixing the error code to the exception message
        try:
            msg = '[{0}] {1}'.format(error_code.name, message.format(*args, **kwargs))
        except (IndexError, KeyError):
            msg = '[{0}] {1}'.format(error_code.name, message)

        super().__init__(msg)


# Error codes for exceptions
class ErrorCodes(Enum, metaclass=EnumMeta):
    pass


class _ErrorCodes(ErrorCodes):
    ERROR_CODE_MISSING = 0  # There is no error code enum specified for the exception
    ERR_INCORRECT_ERRCODE = 1  # error code passed is not specified in enum ErrorCodes
    ERR_INCORRECT_LOGGER = 2  # Logger passed should be of type ESLogger



import os
from typing import Any, Callable, Union
import hvac
from functools import wraps

import requests

from handyman.crypto.exceptions import VaultException, VaultErrorCodes, handle_vault_exceptions
from handyman.crypto import (
    ENCRYPTED_DATA_KEY,
    VAULT_URI,
    VAULT_ROLE_ID,
    VAULT_SECRET_ID,
    VAULT_APPROLE_MOUNTPATH,
    VAULT_DATA_KEY_NAME,
    PLAINTEXT_DATAKEY
)


VAULT_CLIENT = None

# Initialize Vault Client
@handle_vault_exceptions
def vault_client_init():
    """
    Function to initialize Vault Client
    """

    global VAULT_CLIENT

    VAULT_CLIENT = hvac.Client(
        url=VAULT_URI,
    )

# Login with approle auth
@handle_vault_exceptions
def vault_client_login():
    """
    Function to perform vault client login
    """

    global VAULT_CLIENT

    try:
        VAULT_CLIENT.auth.approle.login(
            role_id=VAULT_ROLE_ID,
            secret_id=VAULT_SECRET_ID,
            mount_point=VAULT_APPROLE_MOUNTPATH,
        )
    except requests.exceptions.ConnectionError:
        # Vault connection cant be made
        raise VaultException(VaultErrorCodes.CONNECTION_ERROR, f'Connection with Vault failed, please check if url "{VAULT_URI}" is correct / live')
    except hvac.exceptions.Forbidden:
        # Login failed - login path incorrect / other permission issues
        raise VaultException(VaultErrorCodes.LOGIN_PATH_ERROR, f'Vault login error, check if login path "{VAULT_APPROLE_MOUNTPATH}" is enabled.')
    except (hvac.exceptions.InvalidRequest, hvac.exceptions.InvalidPath):
        # Login failed - role_id or secret_id incorrect
        raise VaultException(VaultErrorCodes.UNAUTHENTICATED, 'Vault login failed, vault credentials incorrect.')
    except Exception as e:
        raise VaultException(VaultErrorCodes.UNKNOWN, 'Unknown error occurred during Vault login') from e

@handle_vault_exceptions
def check_vault_authenticated() -> bool:
    """
    Check if Vault client is authenticated

    :return bool: vault authentication status as boolean
    """

    global VAULT_CLIENT

    is_authenticated = False

    if VAULT_CLIENT.is_authenticated():
        is_authenticated = True

    return is_authenticated

def token_refresh(wrapped: Callable[..., Any]):
    """
    Decorator function to refresh token when wrapped function encounters a code which returns unauthorized from vault
    """

    # Initilize Vault Client if not done earlier
    if not VAULT_CLIENT:
        vault_client_init()

    @wraps(wrapped)
    def wrapper(*args, **kwargs):
        """
        Wrapper function to check for vault unauthorized, refresh token if so
        """

        # Login and get token if no token present
        if check_vault_authenticated():
            vault_client_login()

        try:
            response = wrapped(*args, **kwargs)
        except hvac.exceptions.Forbidden:
            # Retry login in case token expires and re-run wrapped function
            vault_client_login()

            response = wrapped(*args, **kwargs)

        return response

    return wrapper

# Functions for transit keys management

@handle_vault_exceptions
@token_refresh
def transit_key_exist(key_name: str) -> bool:
    """
    Check if transit path exists

    :param str key_name: name of the transit path to check for existence
    :return bool: true is path exists, false if doesn't
    """

    key_exists = False

    try:
        VAULT_CLIENT.secrets.transit.read_key(key_name)

        key_exists = True
    except hvac.exceptions.InvalidPath:
        # Key does not exist
        pass

    return key_exists

@handle_vault_exceptions
@token_refresh
def transit_create_key(key_name: str):
    """
    Create a new transit key (ie a new transit path in vault)

    :param str key_name: name of the transit path to create
    """

    # TBD: allow_plaintext_backup parameter
    try:
        VAULT_CLIENT.secrets.transit.create_key(key_name, key_type='aes256-gcm96')
    except hvac.exceptions.Forbidden:
        # Login failed - login path incorrect / other permission issues
        raise VaultException(VaultErrorCodes.UNAUTHORIZED, f'Vault client unauthorized to create new transit keys, check approle permissions.')

@handle_vault_exceptions
@token_refresh
def transit_delete_key(key_name: str):
    """
    TODO: Delete encryption key from cache after deletion from vault, else encryption will still be available for the client till process restarts
    Delete transit key
    NOTE: Use with caution, once deleted, data key can no longer be decrypted and all encrypted data would be lost

    :param str key_name: name of the transit key to delete
    """

    # If key not present, return
    if not transit_key_exist(key_name):
        return

    # Enable key deletion by updating configuration
    try:
        VAULT_CLIENT.secrets.transit.update_key_configuration(key_name, deletion_allowed=True)
    except hvac.exceptions.Forbidden:
        # Login failed - login path incorrect / other permission issues
        raise VaultException(VaultErrorCodes.UNAUTHORIZED, f'Vault client unauthorized to update transit key configurations, check approle permissions.')

    # Delete the key
    try:
        VAULT_CLIENT.secrets.transit.delete_key(key_name)
    except hvac.exceptions.Forbidden:
        # Login failed - login path incorrect / other permission issues
        raise VaultException(VaultErrorCodes.UNAUTHORIZED, f'Vault client unauthorized to delete transit keys, check approle permissions.')

# Functions for encryption and decryption

@handle_vault_exceptions
@token_refresh
def transit_decrypt_data(transit_path: str, ciphertext: str) -> str:
    """
    Decrypt data at particular transit path.

    :param str transit_path: name of the transit key for decryption
    :param str ciphertext: encrypted data to decrypt
    :return str: decrypted data as plaintext
    """

    global VAULT_CLIENT

    decrypt_data_response = VAULT_CLIENT.secrets.transit.decrypt_data(
        name = transit_path,
        ciphertext = ciphertext,
    )

    if not isinstance(decrypt_data_response, dict):
        raise VaultException(VaultErrorCodes.EMPTY_RESPONSE, "Response from Vault is not a dict")

    try:
        decrypted_data: str = decrypt_data_response["data"]["plaintext"]
    except (KeyError, TypeError):
        raise VaultException(VaultErrorCodes.EMPTY_RESPONSE, "Empty response from Vault while decrypting ciphertext.")

    return decrypted_data

@handle_vault_exceptions
@token_refresh
def transit_generate_data_key(transit_path: str) -> str:
    """
    Generate data key at given transit path.

    :param str transit_path: transit key path
    :return str: encrypted data key as string
    """

    global VAULT_CLIENT

    # Generate a new high entropy data key
    response: dict = VAULT_CLIENT.secrets.transit.generate_data_key(transit_path, "wrapped", bits=256)

    if not isinstance(response, dict):
        raise VaultException(VaultErrorCodes.EMPTY_RESPONSE, "Response from Vault, while generating data key, is not a dict")

    try:
        encrypted_data_key: str = response["data"]["ciphertext"]
    except (KeyError, TypeError):
        raise VaultException(VaultErrorCodes.EMPTY_RESPONSE, "Empty response from Vault, no/incorrect data present in the response of generate data key api of Vault.")

    return encrypted_data_key

@handle_vault_exceptions
@token_refresh
def transit_read_key(transit_path: str) -> Union[dict, None]:
    """
    Read config data at transit path.

    :param str transit_path: transit key path
    :return str: encrypted data key as string
    """

    global VAULT_CLIENT

    try:
        response_: dict = VAULT_CLIENT.secrets.transit.read_key(transit_path)
    except hvac.exceptions.Forbidden:
        # Login failed - login path incorrect / other permission issues
        raise VaultException(VaultErrorCodes.UNAUTHORIZED, f'Vault client unauthorized to read transit keys, check approle permissions.')

    if not isinstance(response_, dict):
        raise VaultException(VaultErrorCodes.EMPTY_RESPONSE, "Response from Vault, while reading transit key configuration, is not a dict")

    try:
        configuration: Union[dict, None] = response_["data"]
    except KeyError:
        raise VaultException(VaultErrorCodes.EMPTY_RESPONSE, "Empty response from Vault, transit configuration data not found")

    return configuration
